'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-08-26 18:35:52
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-08-28 20:35:37
FilePath: /boviz/src/boviz/histogram.py
Description: 
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''



import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter, FixedLocator, NullLocator, NullFormatter

from boviz.config import GLOBAL_COLORS, set_default_dpi_figsize_savedir
from boviz.style import (
    set_default_style,
    set_sans_style,
    set_ax_style,
    apply_axis_scientific_format,
    apply_axis_limits_and_ticks,
    save_or_display_legend,
)
from boviz.utils import generate_plot_filename, save_figure

def plot_histogram(
    label: str,
    x: list[float],
    y: list[float],
    xy_label: tuple[str, str],
    information: str = None,
    factor: list[(float, float)] = None,
    title: str = "Histogram",
    ylog: bool = False,
    sci: tuple[float, float] = [None, None],
    save: bool = False,
    show: bool = False,
    font_style: str = None,
    font_weight: str = "bold",
    categorical_x: bool = True,
    show_values: bool = True,
) -> str:
    """
    绘制柱状图（直方图）
    Args:
        x (list[int]): 核数列表
        y (list[float]): 对应的运行时间列表
        label (str, optional): 图例标签. Defaults to "Runtime".
        xy_label (tuple[str, str], optional): x轴和y轴标签. Defaults to ("Number of Cores", "Runtime (s)").
        title (str, optional): 图表标题. Defaults to "Multi-core Performance".
        legend_location (str, optional): 图例位置. Defaults to "best".
        ylog (bool, optional): 是否使用对数y轴. Defaults to False.
        sci (tuple[float, float], optional): x轴和y轴的科学计数法阈值. Defaults to [None, None].
        save (bool, optional): 是否保存图表. Defaults to True.
        show (bool, optional): 是否显示图表. Defaults to True.
        font_style (str, optional): 字体风格，"sans"表示无衬线字体，None表示默认字体. Defaults to None.
        font_weight (str, optional): 字体粗细，"bold"表示粗体，"normal"表示正常字体. Defaults to "bold".
        categorical_x (bool, optional): 是否将x轴作为类别处理（等间距）. Defaults to True.
    Returns:
        str: 保存的图表路径
    """


    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")
        
    dpi, figuresize, savedir = set_default_dpi_figsize_savedir()
    fig, ax = plt.subplots(figsize=figuresize, dpi=dpi)
    save_dir = os.path.join(savedir, "Histograms")

    set_ax_style(ax)
    ax.tick_params(axis="both", direction="in", width=3, which="both", pad=10)
    
    if factor:
        y = np.asarray(y, dtype=float)
        y = y * factor[0] + factor[1]

    # 绘制柱状图
    colors = plt.cm.Dark2.colors
    if categorical_x:
        width = 0.28
        x_pos = np.arange(len(x))
        bars = ax.bar(x_pos, y, color=colors, edgecolor="black", width=width, label=label)
        ax.set_xticks(x_pos)
        ax.set_xticklabels([str(v) for v in x], fontweight=font_weight)

        # ✅ 两端空白 = 条间距的一半
        gap = 0.75
        ax.set_xlim(-gap, len(x) - 1 + gap)
    else:
        # 数值轴：按数值比例排布（你原来的画法）
        bars = ax.bar(x, y, color=colors, edgecolor="black", width=0.6, label=label)
        ax.set_xlim(min(x) - 0.6, max(x) + 0.6)

    if show_values:
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax * 1.1)   # 上限增加 5%
        for bar, val in zip(bars, y):
            ax.text(
                bar.get_x() + bar.get_width() / 2,
                bar.get_height(),
                f"{val:.2f}",    # 保留两位小数
                ha="center", va="bottom",
                fontweight=font_weight,
                fontsize=24
            )

    ax.set_title(title, pad=20, fontweight=font_weight)
    ax.set_xlabel(xy_label[0], fontweight=font_weight)
    ax.set_ylabel(xy_label[1], fontweight=font_weight)

    label_suffix = f"({information})" if information else None

    if sci[0]:
        apply_axis_scientific_format(ax, 'x', sci[0])
    if sci[1]:
        apply_axis_scientific_format(ax, 'y', sci[1])

    if ylog:
        ax.set_yscale('log')
        ax.set_ylim(0.0, 0.6)
        ax.yaxis.set_major_locator(FixedLocator([0.002, 0.01, 0.1, 0.6]))
        ax.yaxis.set_major_formatter(FuncFormatter(lambda y, _: f'{y:g}'))
        ax.yaxis.set_minor_locator(NullLocator())
        ax.yaxis.set_minor_formatter(NullFormatter())

    plt.tight_layout()
    filename = generate_plot_filename(title=title, suffix=label_suffix)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close()
    return save_path





'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 16:58:46
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-09-01 15:39:54
FilePath: /boviz/src/boviz/utils.py
Description: This module provides utility functions for boviz, including generating standardized plot filenames.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''


import os
import numpy as np
import pandas as pd
import meshio
from datetime import datetime
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from netCDF4 import chartostring

def generate_plot_filename(title: str, suffix=None) -> str:
    """
    生成统一命名格式的图片文件名，格式为：boviz_YYMMDDHHMM_title_suffix.png

    Args:
        title (str): 图像标题或描述性名称（可含空格，会被自动替换为下划线）。
        suffix (str, optional): 附加信息（如 "(test)"），默认空字符串。

    Returns:
        str: 构造后的图片文件名（不含路径）。
    """
    timestamp = datetime.now().strftime("%y%m%d%H%M")
    title_clean = title.replace(" ", "") if title else "plot"
    if suffix is None:
        return f"boviz_{timestamp}_{title_clean}.png"
    else:
        return f"boviz_{timestamp}_{title_clean}{suffix}.png"


def save_figure(save_path: str, dpi: int = 300, verbose: bool = True):
    """
    保存当前图像到指定路径，并确保目录存在。

    Args:
        save_path (str): 图像保存完整路径（含文件名）。
        dpi (int): 图像分辨率，默认 300。
        verbose (bool): 是否打印保存信息，默认 True。
    """
    os.makedirs(os.path.dirname(save_path), exist_ok=True)
    plt.savefig(save_path, dpi=dpi, bbox_inches='tight')
    if verbose:
        print(f"[SAVE] 图像已保存到: {save_path}\n")


def load_data_csv(
    source: str,
    x_index: int,
    y_index: int,
    factor: tuple[tuple, tuple],
    time_step: int = 0
) -> tuple[np.ndarray, np.ndarray, str, str]:
    """
    从 CSV 文件中读取指定列的数据。

    Args:
        source (str): CSV 文件路径。
        x_index (int): X 轴数据的列索引。
        y_index (int): Y 轴数据的列索引。
        factor (tuple[touple, touple]): 用于缩放和平移 Y 轴数据的因子，格式为 [scale, offset]。
            - scale: 缩放因子，默认为 1.0。
            - offset: 平移量，默认为 0.0。
        time_step (int): 若不为 0，则只保留前 time_step 个时间步。

    Returns:
        x_data, y_data: 对应列的数据（NumPy 数组）
        x_colname, y_colname: 对应列名
    """

    if factor is None:
        factor = [(1.0, 0.0), (1.0, 0.0)]

    if not isinstance(source, str) or not source.endswith('.csv'):
        raise ValueError(f"Expected a .csv file path, got {source}")

    df = pd.read_csv(source)
    x_data_raw = df.iloc[:time_step, x_index] if time_step else df.iloc[:, x_index]
    y_data_raw = df.iloc[:time_step, y_index] if time_step else df.iloc[:, y_index]
    x_colname = df.columns[x_index]
    y_colname = df.columns[y_index]

    x_data = x_data_raw * factor[0][0] + factor[0][1]
    y_data = y_data_raw * factor[1][0] + factor[1][1]

    return x_data.values, y_data.values, x_colname, y_colname

def generate_particle_layout(
        num_x: int, 
        num_y: int,
        radius: float,
        border: float = None,
    ) -> np.ndarray:
    """
    生成粒子布局的网格坐标。

    Args:
        num_x (int): X 方向的粒子数量。
        num_y (int): Y 方向的粒子数量。
        radius (float): 粒子的半径。
        border (float, optional): 边界的宽度，默认为2倍颗粒半径。

    Returns:
        tuple: 包含三个元素的元组：
            - centers_coordinate (list): 粒子中心坐标的列表，每个元素为 [x, y]。
            - radii (list): 每个粒子的半径列表。
            - domain_size (list): 网格的域大小 [domain_x, domain_y]。
    """
    if border is None:
        border = 2.0
    border = border * radius
    domain_x = 2 * radius * num_x + border * 2
    domain_y = 2 * radius * num_y + border * 2
    
    radii = [radius] * (num_x * num_y)
    centers_coordinate = []
    for j in range(num_y):
        for i in range(num_x):
            x_coordinate = int(domain_x / 2 + (i + (1 - num_x) / 2) * radius * 2)
            y_coordinate = int(domain_y / 2 + (j + (1 - num_y) / 2) * radius * 2)
            centers_coordinate.append([x_coordinate, y_coordinate])
    
    domain_size = [domain_x, domain_y]
    return centers_coordinate, radii, domain_size

def build_tanh_phase_field(
        centers_coordinate: list,
        radii: list,
        domain_size: list,
        tanh_width: float = 3.0,
        tanh_offset: float = 0.05
    ) -> np.ndarray:
    """
    构建基于双曲正切函数的相场。

    Args:
        centers_coordinate (list): 粒子中心坐标列表，每个元素为 [x, y]。
        radii (list): 每个粒子的半径列表。
        domain_size (list): 网格的域大小 [domain_x, domain_y]。
        tanh_width (float): 双曲正切函数的宽度，默认值为 3.0。
        tanh_offset (float): 双曲正切函数的偏移量，默认值为 0.05。

    Returns:
        np.ndarray: 生成的相场数组，大小为 [domain_x * 10, domain_y * 10]。
    """
    domain_x, domain_y = domain_size
    phase_field = 0
    x, y = np.meshgrid(np.arange(0, domain_x+0.1, 0.1), np.arange(0, domain_y+0.1, 0.1))
    for center, radius in zip(centers_coordinate, radii):
        distance = np.sqrt((x - center[0])**2 + (y - center[1])**2)
        phase_field += 0.5 * (1 - np.tanh((distance - radius) * (2 * np.arctanh(1 - 2 * tanh_offset)) / tanh_width))
    
    return phase_field


def load_exodus_data(
    source: str,
    variable_name: str,
    time_step: int = 0,
):
    """
    读取 Exodus 文件中的指定变量的指定时间步数据。
    
    args:
        source (str): Exodus 文件路径。
        variable_name (str): 要读取的变量名称。
        time_step (int): 时间步索引，默认为 0。
    returns:
        coordinates (np.ndarray): 网格节点坐标数组，形状为 (N, 2)，其中 N 是节点数量。
        variable_values (np.ndarray): 指定变量在指定时间步的值，形状为 (N,) 或 (N, M)，其中 M 是变量的维度（如标量或向量）。
    raises:
        ValueError: 如果指定的变量在 Exodus 文件中不存在。
    raises:
        FileNotFoundError: 如果指定的 Exodus 文件不存在。
    """

    print(meshio.__file__)   # 输出模块位置
    print(meshio.read)       # 输出 read 函数指针

    mesh = meshio.read(source, time_step=time_step)
    coordinates = mesh.points



    try:
        variable_values = mesh.point_data[variable_name]
        if variable_values.ndim == 2:
            variable_values = variable_values[time_step]
    except KeyError:
        raise ValueError(f"Variable '{variable_name}' not found in the Exodus file.")
    
    return coordinates, variable_values


def load_exodus_data_netcdf(source, variable_name, time_step=0):
    with Dataset(source, 'r') as f:
        # 获取节点坐标
        coords_x = f.variables["coordx"][:]
        coords_y = f.variables["coordy"][:]
        coordinates = np.stack([coords_x, coords_y], axis=1)

        # 获取变量名表
        raw_names = f.variables["name_nod_var"][:]
        var_names = chartostring(raw_names).tolist()
        var_names = [n.strip() for n in var_names]

        # 读取时间信息
        time_array = f.variables["time_whole"][:]
        num_steps = len(time_array)
        t = time_array[time_step]

        # 打印美观输出
        print("[INFO] Exodus 数据加载信息：")
        print(f"   ├─ 文件路径：{source}")
        print("   ├─ 可用变量：", var_names)
        print(f"   ├─ 当前变量：'{variable_name}'")
        print(f"   ├─ 当前时间步：第 {time_step} 步（时间 = {time_array[time_step]:.4f}）")
        print(f"   ├─ 总时间步数：{num_steps}")
        preview_steps = ", ".join([f"{t:.4f}" for t in time_array[:min(5, num_steps)]])
        tail = f", ..., {time_array[-1]:.4f}" if num_steps > 5 else ""
        print(f"   └─ 时间序列预览：[ {preview_steps}{tail} ]")

        if variable_name not in var_names:
            raise ValueError(f"变量 '{variable_name}' 不存在，可选项: {var_names}")

        print(f"[INFO] 正在绘制热图...请稍候...")
        idx = var_names.index(variable_name) + 1
        variable_values = f.variables[f"vals_nod_var{idx}"][time_step, :]
    
    # 自动生成图标题
    math_label = get_math_label(variable_name)
    title = f"{math_label} at {t:.4g}s"
    save_name = f"{variable_name}_{t:.4g}"

    return coordinates, variable_values, title, save_name


def get_math_label(var_name: str) -> str:
    """
    将变量名称映射为 LaTeX 数学表达式字符串。
    """
    mapping = {
        "F_density": r"$f$",
        "Real_Pressure": r"$P_\mathrm{real}$",
        "Stress_Magnitude": r"$|\boldsymbol{\sigma}|$",
        "Stress_xx": r"$\sigma_{xx}$",
        "Stress_xy": r"$\sigma_{xy}$",
        "Stress_yx": r"$\sigma_{yx}$",
        "Stress_yy": r"$\sigma_{yy}$",
        "V_Magnitude": r"$|\boldsymbol{v}|$",
        "c": r"$C$",
        "mu": r"$\mu$",
        "p": r"$p$",
        "u": r"$u_x$",
        "v": r"$u_y$",
    }
    return mapping.get(var_name, var_name)


def _broadcast(param, N):
    if isinstance(param, (list, tuple, np.ndarray)):
        if len(param) != N:
            raise ValueError(f"参数长度应为 {N}，但收到 {len(param)}")
        return list(param)
    return [param] * N

def _bbox_cols_from_gridspec(fig, gs, nx, ny):
    """取前 nx 列（跨 ny 行）的列槽位左右范围（figure 坐标）。"""
    fig.canvas.draw()
    # 用第一行的前 nx 列即可得到列左右边界
    bb_row0 = gs[0, :nx].get_position(fig)
    return bb_row0.x0, bb_row0.x1

def _bbox_rows_from_axes(fig, axes):
    """取实际 Axes（不含面板标题那段负偏移文本）的上下范围（figure 坐标）。"""
    fig.canvas.draw()
    y0 = min(ax.get_position().y0 for ax in axes)
    y1 = max(ax.get_position().y1 for ax in axes)
    return y0, y1



'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 15:28:18
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-08-28 20:25:38
FilePath: /boviz/src/boviz/config.py
Description: This module defines global configuration settings for boviz, including default colors, save directory, DPI, and figure size.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''



import os

# 全局颜色列表（可自定义扩展）
GLOBAL_COLORS = [
    'tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple',
    'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan', 'black',
    '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
    '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'
]

def set_default_dpi_figsize_savedir():
    """
    Set default DPI, figure size, and save directory for plots.

    Args:
        None

    Returns:
        tuple: A tuple containing default DPI, figure size, and save directory.
    """
    savedir = os.path.join(os.getcwd(), "figures")
    os.makedirs(savedir, exist_ok=True)
    default_dpi = 100
    default_figsize = (12, 9)
    return default_dpi, default_figsize, savedir

def set_residual_dpi_figsize_savedir():
    """
    Set default DPI, figure size, and save directory for residual plots.

    Args:
        None

    Returns:
        tuple: A tuple containing default DPI, figure size, and save directory.
    """
    savedir = os.path.join(os.getcwd(), "figures")
    os.makedirs(savedir, exist_ok=True)
    default_dpi = 100
    default_figsize = (12, 9)
    return default_dpi, default_figsize, savedir


'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 17:14:02
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-08-18 16:36:39
FilePath: /boviz/src/boviz/schematic.py
Description: This module provides a function to plot the initial distribution of particles in a schematic format, including their positions and radii.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''



import os
import numpy as np
import matplotlib.pyplot as plt

from boviz.config import set_default_dpi_figsize_savedir
from boviz.style import set_default_style, set_ax_style, set_sans_style, set_smart_xy_ticks
from boviz.utils import generate_plot_filename, save_figure

def plot_initial_particle_schematic(
    coordinates: list,
    radii: list,
    domain: list,
    title: str = "Initial Particle Distribution",
    show: bool = False,
    save: bool = False,
    font_style: str = None,
    font_weight: str = "bold",
):
    """
    绘制初始粒子分布的示意图
    Args:
        coordinates (list): 粒子中心坐标列表，格式为 [[x1, y1], [x2, y2], ...]。
        radii (list): 粒子半径列表，格式为 [r1, r2, ...]。
        domain (list): 绘图区域的空间大小，格式为 [width, height]。
        title (str): 图表标题。
        show (bool): 是否显示图像，默认不显示。
        save (bool): 是否保存图像，默认不保存。
        font_style (str): 字体样式，默认为 Times。可选值为 'sans' 或 None。
        font_weight (str): 字体粗细，默认为 "bold"。可选值为 'bold' 或 'normal'.
    """
    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    save_dir = os.path.join(set_default_dpi_figsize_savedir()[2], "InitialSchematic")

    filename = generate_plot_filename(title=title)
    save_path = os.path.join(save_dir, filename)
    
    fig, ax = plt.subplots(figsize=set_default_dpi_figsize_savedir()[1], dpi=set_default_dpi_figsize_savedir()[0])
    set_ax_style(ax)

    set_smart_xy_ticks(ax, extent=(0, domain[0], 0, domain[1]))
    # ax.set_xlim(0, domain[0])
    # ax.set_ylim(0, domain[1])
    # ax.set_xticks(np.arange(0, domain[0] + 1, 30))
    # ax.set_yticks(np.arange(0, domain[1] + 1, 30))
    ax.set_aspect('equal', 'box')

    for i in range(len(coordinates)):
        circle = plt.Circle(
            (coordinates[i][0], coordinates[i][1]),
            radii[i],
            edgecolor='black',
            facecolor='white',
            linewidth=3,
            zorder=2
        )
        ax.add_artist(circle)
        plt.text(
            coordinates[i][0],
            coordinates[i][1],
            rf"$\text{{Particle}}_{{{i + 1}}}$",
            fontsize=32,
            ha='center',
            va='center',
            zorder=3
        )

    ax.grid(True, linestyle='--', linewidth=3, zorder=1)
    plt.tick_params(axis='both', direction='in', width=3, which='both', pad=10)
    plt.xlabel('X-axis', fontweight=font_weight)
    plt.ylabel('Y-axis', fontweight=font_weight)
    plt.title(title, pad=20, fontweight=font_weight)

    plt.tight_layout()
    if save:
        save_figure(save_path, dpi=set_default_dpi_figsize_savedir()[0])
    if show:
        plt.show()
    plt.close()
    return save_path



'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-29 13:54:52
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-09-01 15:42:44
FilePath: /boviz/src/boviz/heatmap.py
Description: Plotting module for generating heatmaps of particle distributions.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''


import os
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable, axes_size
from matplotlib import gridspec
from matplotlib.colors import Normalize
from matplotlib.cm import ScalarMappable

from boviz.style import set_default_style, set_smart_xy_ticks, set_sans_style
from boviz.config import set_default_dpi_figsize_savedir
from boviz.utils import generate_plot_filename, generate_particle_layout, build_tanh_phase_field, save_figure, load_exodus_data_netcdf, _broadcast, _bbox_cols_from_gridspec, _bbox_rows_from_axes


def plot_heatmap_particle(
    particle_x_num: int,
    particle_y_num: int,
    particle_radius: float,
    border: float = None,
    cmap: str = 'coolwarm',
    title_figure: str = "Initial Particle Schematic",
    show: bool = False,
    save: bool = False,
    information: str = None,
    surface_thickness: float = 3.0,
    tanh_offset: float = 0.05,
    font_style: str = None,
    font_weight: str = "bold",
    show_ticks: bool = True,
):
    """
    绘制初始粒子分布的热图。

    Args:
        particle_x_num (int): 粒子在x方向的数量。
        particle_y_num (int): 粒子在y方向的数量。
        particle_radius (float): 粒子的半径。
        border (float, optional): 粒子布局的边界宽度，默认为 None。
        cmap (str, optional): 热图使用的颜色映射，默认为 'coolwarm', 可选值包括 'viridis', 'plasma', 'inferno', 'magma', 'cividis' 等。
        title_figure (str, optional): 图像标题，默认为 "Initial Particle Schematic"。
        show (bool, optional): 是否显示图像，默认为 True。
        save (bool, optional): 是否保存图像，默认为 True。
        information (str, optional): 附加信息，用于生成文件名后缀。
        surface_thickness (float, optional): 表面厚度，用于生成相场，默认为 3.0。
        tanh_offset (float, optional): 相场的偏移量，默认为 0.05。
        font_style (str, optional): 字体样式，默认为 Times。可选值为 'sans' 或 None。
        font_weight (str, optional): 字体粗细，默认为 None。可选值为 'bold' 或 None。
        show_ticks (bool, optional): 是否显示坐标轴刻度以及标题，默认为 True。

    Returns:
        str: 保存的图像路径。
    """
    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    dpi, figuresize, savedir = set_default_dpi_figsize_savedir()
    fig, ax = plt.subplots(figsize=figuresize, dpi=dpi)
    save_dir = os.path.join(savedir, "HeatMaps")
    label_suffix = f"({information})" if information else None

    particle_center_coordinate, radii, domain_size = generate_particle_layout(
        particle_x_num, particle_y_num, particle_radius, border=border
    )
    phase_field = build_tanh_phase_field(
        centers_coordinate=particle_center_coordinate,
        radii=radii,
        domain_size=domain_size,
        tanh_width=surface_thickness,
        tanh_offset=tanh_offset
    )

    heatmap = ax.imshow(
        phase_field,
        cmap=cmap,
        extent=[0, domain_size[0], 0, domain_size[1]],
        origin='lower',
        aspect='auto'
    )

    if show_ticks:
        set_smart_xy_ticks(ax)
        ax.set_xlabel('X Coordinate', fontweight=font_weight)
        ax.set_ylabel('Y Coordinate', fontweight=font_weight)
        ax.set_title(title_figure, pad=20, fontweight=font_weight)
    else:
        ax.set_xticks([])
        ax.set_yticks([])
    ax.set_aspect('equal', adjustable='box')
    divider = make_axes_locatable(plt.gca())
    width = axes_size.AxesY(ax, aspect=1. / 20)
    pad = axes_size.Fraction(0.5, width)
    cax = divider.append_axes("right", size=width, pad=pad)
    cbar = plt.colorbar(heatmap, cax=cax)

    plt.tight_layout()
    filename = generate_plot_filename(title=title_figure, suffix=label_suffix)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close()
    return save_path


def plot_heatmap_exodus2d(
    path: str,
    variable: str,
    colorbar_range: tuple = None,
    time_step: int = 0,
    cmap: str = 'coolwarm',
    title_figure: str = None,
    show: bool = False,
    save: bool = False,
    information: str = None,
    font_style: str = None,
    font_weight: str = "bold",
    show_ticks: bool = True,
):
    """
    绘制 Exodus 2D 数据的热图。

    Args:
        None

    Returns:
        str: 保存的图像路径。
    """
    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    dpi, figuresize, savedir = set_default_dpi_figsize_savedir()
    fig, ax = plt.subplots(figsize=figuresize, dpi=dpi)
    save_dir = os.path.join(savedir, "HeatMaps")
    label_suffix = f"({information})" if information else None
    
    coordinates, variable_values, title, save_name = load_exodus_data_netcdf(
        source=path, 
        variable_name=variable, 
        time_step=time_step
    )

    x = coordinates[:, 0]
    y = coordinates[:, 1]

    if colorbar_range is None:
        heatmap = ax.tricontourf(x, y, variable_values, cmap=cmap, levels=256, origin='lower')
        vmin, vmax = heatmap.get_clim()
    else:
        vmin, vmax = colorbar_range
        levels = np.linspace(vmin, vmax, 256)
        heatmap = ax.tricontourf(x, y, variable_values, cmap=cmap, levels=levels, origin='lower')

    print(f"[INFO] 热图绘制完成。")

    if show_ticks:
        set_smart_xy_ticks(ax)
        ax.set_xlabel('X Coordinate', fontweight=font_weight)
        ax.set_ylabel('Y Coordinate', fontweight=font_weight)
    else:
        ax.set_xticks([])
        ax.set_yticks([])
    
    if title_figure is None:
        title_figure = title
    ax.set_title(title_figure, pad=20, fontweight=font_weight)
    ax.set_aspect('equal', adjustable='box')
    divider = make_axes_locatable(plt.gca())
    width = axes_size.AxesY(ax, aspect=1. / 20)
    pad = axes_size.Fraction(0.5, width)
    cax = divider.append_axes("right", size=width, pad=pad)
    cbar = plt.colorbar(heatmap, cax=cax)
    
    tick_locs = np.linspace(vmin, vmax, 6)
    cbar.set_ticks(tick_locs)
    cbar.set_ticklabels([f"{v:.1f}".replace("-0.0", "0.0") for v in tick_locs])


    plt.tight_layout()
    filename = generate_plot_filename(title=save_name, suffix=label_suffix)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close()
    return save_path


def plot_heatmap_exodus2d_grid(
    nx: int,
    ny: int,
    paths,
    variables,
    time_steps=0,
    cmap='coolwarm',
    colorbar_range: tuple | None = None,   # (vmin, vmax)；None 自动
    titles: list | None = None,            # 每个子图标题；None 用文件内标题
    suptitle: str | None = None,           # 总标题
    panel_title_size: int = 20,            # 小图标题字号（更小）
    suptitle_size: int = 24,               # 总标题字号
    cbar_width: float = 0.025,     # ← 色条宽度（相对图像坐标）
    cbar_pad: float = 0.025,      # ← 色条与右侧子图之间的间距
    information: str | None = None,
    font_style: str | None = None,
    font_weight: str = 'bold',
    show: bool = False,
    save: bool = False,
):
    """
    绘制 Exodus2D 网格数据的热力图（支持多子图排版）。

    Args:
        nx : int
            子图网格的列数（每行的子图数量）。
        ny : int
            子图网格的行数（每列的子图数量）。
        paths : str 或 list
            Exodus2D NetCDF 文件路径，单个或多个（会自动广播到 nx*ny 个）。
        variables : str 或 list
            需要绘制的变量名，单个或多个（会自动广播到 nx*ny 个）。
        time_steps : int 或 list, 默认 0
            变量的时间步索引，单个或多个（会自动广播到 nx*ny 个）。
        cmap : str 或 list, 默认 'coolwarm'
            色图（colormap），单个或多个（会自动广播到 nx*ny 个）。
        colorbar_range : tuple(vmin, vmax) 或 None, 默认 None
            色条的取值范围 (vmin, vmax)。若为 None，则自动根据所有数据计算。
        titles : list 或 None, 默认 None
            每个子图的标题。若为 None，则使用文件内自带标题。
        suptitle : str 或 None, 默认 None
            总标题，显示在所有子图上方。
        panel_title_size : int, 默认 20
            每个子图标题的字体大小。
        suptitle_size : int, 默认 24
            总标题的字体大小。
        cbar_width : float, 默认 0.025
            色条宽度（相对于整个 figure 的宽度）。
        cbar_pad : float, 默认 0.025
            色条与右侧子图之间的间距（相对于整个 figure 的宽度）。
        information : str 或 None, 默认 None
            附加信息，会作为文件名后缀。
        font_style : str 或 None, 默认 None
            字体风格。'sans' 表示无衬线字体，None 表示默认字体。
        font_weight : str, 默认 'bold'
            字体粗细。'bold' 或 'normal'。
        show : bool, 默认 False
            是否显示图像（plt.show）。
        save : bool, 默认 False
            是否保存图像到文件。

    Returns:
        save_path : str
            图像保存路径（若 save=True），否则返回最后一次保存路径。
    """
    # 字体与风格
    if not font_style:
        if font_weight == 'bold': set_default_style(bold=True)
        elif font_weight == 'normal': set_default_style(bold=False)
        else: raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold': set_sans_style(bold=True)
        elif font_weight == 'normal': set_sans_style(bold=False)
        else: raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    N = nx * ny
    paths      = _broadcast(paths, N)
    variables  = _broadcast(variables, N)
    time_steps = _broadcast(time_steps, N)
    cmaps      = _broadcast(cmap, N)

    dpi, figuresize, savedir = set_default_dpi_figsize_savedir()

    fig = plt.figure(figsize=figuresize, dpi=dpi)
    gs = gridspec.GridSpec(
        nrows=ny, ncols=nx + 1,
        width_ratios=[1] * nx + [nx/(22-nx)],   # 右侧留出 colorbar 列
        height_ratios=[1] * ny,
        figure=fig
    )

    # 载入数据，计算全局色标范围
    recs = []  # (x, y, vals, default_title, save_name)
    gmin, gmax = np.inf, -np.inf
    for i in range(N):
        print(f"\n获取第 {i+1} 张图的数据...")
        coords, vals, dtitle, sname = load_exodus_data_netcdf(
            source=paths[i], variable_name=variables[i], time_step=time_steps[i]
        )
        x, y = coords[:, 0], coords[:, 1]
        recs.append((x, y, vals, dtitle, sname))
        if colorbar_range is None and np.size(vals):
            vmin_i, vmax_i = np.nanmin(vals), np.nanmax(vals)
            gmin = min(gmin, vmin_i); gmax = max(gmax, vmax_i)

    if colorbar_range is None:
        vmin, vmax = gmin, gmax
    else:
        vmin, vmax = colorbar_range
    if not np.isfinite(vmin) or not np.isfinite(vmax):
        raise ValueError("无法确定 colorbar 范围（数据为空或全为 NaN）。")

    # 统一色标
    norm = Normalize(vmin=vmin, vmax=vmax)
    cmap_common = cmaps[0]   # 共享色条需单一 cmap；如需每图不同色表，就不能共享色条
    sm = ScalarMappable(norm=norm, cmap=cmap_common)
    sm.set_array([])

    print("\n正在排版...")
    axes = []
    for k in range(N):
        r, c = divmod(k, nx)
        ax = fig.add_subplot(gs[r, c])
        x, y, vals, dtitle, sname = recs[k]

        levels = np.linspace(vmin, vmax, 256)
        m = ax.tricontourf(x, y, vals, levels=levels, cmap=cmaps[k], origin='lower')
        ax.set_xticks([]); ax.set_yticks([])

        # 小图标题放在图下方居中
        t = dtitle if titles is None else _broadcast(titles, N)[k]
        ax.text(0.5, -0.04, t, ha='center', va='top', transform=ax.transAxes,
                fontsize=panel_title_size, fontweight=font_weight, clip_on=False)

        ax.set_aspect('equal', adjustable='box')
        axes.append(ax)

    # x 边界：用 GridSpec 列槽位；y 边界：用实际 Axes
    fig.subplots_adjust(
        wspace=nx/(22-nx),   # 子图之间的水平间距（默认大约0.2）
        hspace=ny/(22-ny)    # 子图之间的垂直间距（默认大约0.2）
    )   
    left, right = _bbox_cols_from_gridspec(fig, gs, nx, ny)
    bottom, top = _bbox_rows_from_axes(fig, axes)
    host = fig.add_axes([left, bottom, right - left, top - bottom], frameon=False)
    host.set(xticks=[], yticks=[])
    for sp in host.spines.values():
        sp.set_visible(False)
    # 仿单图：用 divider 在右侧追加色条轴
    pad_fig = cbar_pad   # 右侧留白（占 figure 宽度），可微调
    w_fig   = cbar_width  # 色条宽度（占 figure 宽度），对应你原 cbar_width

    cax = fig.add_axes([right + pad_fig, bottom, w_fig, top - bottom])

    cbar  = fig.colorbar(sm, cax=cax)
    ticks = np.linspace(vmin, vmax, 6)
    cbar.set_ticks(ticks)
    cbar.set_ticklabels([f"{v:.1f}".replace("-0.0", "0.0") for v in ticks])
    cbar.outline.set_visible(True)

    # 总标题置顶居中
    if suptitle:
        fig.suptitle(suptitle, fontsize=suptitle_size, fontweight=font_weight, y=0.99)

    # 保存/显示
    label_suffix = f"({information})" if information else None
    filename = generate_plot_filename(title="Exodus2D_Grid", suffix=label_suffix)
    save_dir = os.path.join(savedir, "HeatMaps"); os.makedirs(save_dir, exist_ok=True)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close(fig)
    return save_path


'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-26 15:45:34
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-06-28 01:39:40
FilePath: /boviz/tests/test_plot_curves.py
Description: This module contains unit tests for the boviz plotting functions, ensuring they work correctly with sample data.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''
import os
import pytest
import pandas as pd
import matplotlib.pyplot as plt
from boviz import plot_curves_csv, plot_curves, plot_initial_particle_schematic, generate_plot_filename

# 测试数据路径（假设你提供了样例 csv）
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
TEST_CSV = os.path.join(DATA_DIR, "sample_curve.csv")

@pytest.fixture
def setup_sample_csv(tmp_path):
    # 创建一个简单的 csv 文件
    file_path = tmp_path / "sample_curve.csv"
    df = pd.DataFrame({
        "Time": [0, 1, 2, 3, 4, 5],
        "Value1": [0.0, 0.5, 0.9, 1.4, 1.6, 2.0],
        "Value2": [0.0, 0.4, 0.8, 1.3, 1.5, 1.9]
    })
    df.to_csv(file_path, index=False)
    return str(file_path)

def test_plot_single_curve(setup_sample_csv):
    save_path = plot_curves_csv(
        path=[setup_sample_csv],
        label=["Test Curve"],
        x=[0],
        y=[1],
        title_figure="Single Curve Plot",
        information="test_case_single",
        save=True,
        show=False
    )
    assert os.path.exists(save_path)

def test_particle_schematic():
    coordinates = [[10, 10], [20, 20]]
    radii = [2, 3]
    domain = [30, 30]
    save_path = plot_initial_particle_schematic(
        coordinates, radii, domain,
        title="InitialParticleDistribution",
        save=True,
        show=False
    )
    assert os.path.exists(save_path)




'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 19:27:14
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-09-01 15:58:29
FilePath: /boviz/setup.py
Description: This script sets up the boviz package for distribution, including metadata and dependencies.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''
from setuptools import setup, find_packages

setup(
    name="boviz",
    version="0.2.3",
    author="Bo Qian",
    author_email="bqian@shu.edu.cn",
    description="Bo Qian's advanced scientific plotting toolkit",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://github.com/bo-qian/boviz",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas",
        "pytest",
        "pathlib",
        "argparse",
        "meshio>=4.0",
        "netCDF4>=1.5"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Visualization"
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'boviz = boviz.cli:main',
        ],
    },
)


'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 15:38:39
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-08-28 20:45:37
FilePath: /boviz/src/boviz/curves.py
Description: This module provides functions to plot curves with various styles and options, including support for multiple curves, residual analysis, and custom styling.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import ticker
from matplotlib.ticker import FuncFormatter, FixedLocator, NullLocator, NullFormatter

from boviz.config import GLOBAL_COLORS, set_default_dpi_figsize_savedir, set_residual_dpi_figsize_savedir
from boviz.style import set_default_style, set_ax_style, apply_axis_scientific_format, apply_axis_limits_and_ticks, save_or_display_legend, plot_residual_curves, set_sans_style
from boviz.utils import generate_plot_filename, load_data_csv, save_figure

def update_curve_plotting_with_styles(ax, x_data, y_data, label, index):
    line_styles = ['-', '--', '-.', ':']
    markers = ['o', 's', 'D', '^', 'v', '*']
    color = GLOBAL_COLORS[index % len(GLOBAL_COLORS)]

    ax.plot(x_data, y_data,
            label=label,
            linestyle=line_styles[index % len(line_styles)],
            marker=markers[index % len(markers)],
            markevery=slice(index * 2, None, max(1, len(x_data) // 20)),
            markersize=6,
            linewidth=3 if index == 0 else 2,
            color=color,
            alpha=0.9)


def plot_scatter_style(ax, x_data, y_data, label, index):
    markers = ['o', 's', 'D', '^', 'v', '*']
    color = GLOBAL_COLORS[index % len(GLOBAL_COLORS)]

    ax.scatter(x_data, y_data,
               label=label,
               s=60,
               marker=markers[index % len(markers)],
               edgecolors=color,
               facecolors=color,
               linewidths=1.5,
               zorder=5)


def plot_curves_csv(
    path: list[str],
    label: list[str],
    x: list[int],
    y: list[int],
    information: str = None,
    factor: list[tuple[tuple, tuple]] = None,
    time_step: list[int] = None,
    xy_label: tuple[str, str] = None,
    use_marker: list[bool] = None,
    use_scatter: list[bool] = None,
    tick_interval_x: float = None,
    tick_interval_y: float = None,
    legend_location: str = None,
    xlim: tuple[float, float] = None,
    ylim: tuple[float, float] = None,
    highlight_x: float = None,
    split_legend: bool = False,
    show_residual: bool = False,
    show_legend: bool = True,
    title_figure: str = None,
    legend_ncol: int = None,
    ylog: bool = False,
    sci: tuple[float, float] = [None, None],
    color_group: list[int] = None,
    show: bool = False,
    save: bool = False,
    font_style: str = None,
    font_weight: str = "bold",
) -> str:
    """
    绘制一条或多条曲线，支持多种样式控制与残差分析的通用函数。

    Args:
        path (List[str]): 每条曲线对应的 CSV 文件路径。
        label (List[str]): 每条曲线的图例标签。
        x (List[int]): 每条曲线 X 数据所在列的索引。
        y (List[int]): 每条曲线 Y 数据所在列的索引。
        information (str, optional): 文件名后缀信息，用于区分保存图像。
        factor (List[tuple[tuple, tuple]]], optional): x轴和y轴的缩放因子和偏移量，例如用于单位换算。格式为 [((x_scale, x_offset), (y_scale, y_offset)), ...]。默认均为 1 和 0。
        time_step (List[int], optional): 用于截断数据的步数，若为 0 表示不截断。
        xy_label (Tuple[str, str], optional): X 和 Y 轴标签，例如 ("Time (s)", "Stress (MPa)")。
        use_marker (List[bool], optional): 是否为每条曲线使用线+标记风格。
        use_scatter (List[bool], optional): 是否将每条曲线绘制为散点图。
        tick_interval_x (float, optional): X 轴主刻度间隔。
        tick_interval_y (float, optional): Y 轴主刻度间隔。
        legend_location (str, optional): 图例位置（如 'best'、'upper right' 等）。
        xlim (Tuple[float, float], optional): X 轴显示范围。
        ylim (Tuple[float, float], optional): Y 轴显示范围。
        highlight_x (float, optional): 保留接口，尚未启用，用于高亮指定 X 点。
        split_legend (bool, optional): 是否将图例单独绘制成图像保存。
        show_residual (bool, optional): 是否绘制与参考曲线（第 1 条）相比的残差图。
        title_figure (str, optional): 图像标题，也作为保存文件名的前缀。
        legend_ncol (int, optional): 图例列数，默认自动调整。
        ylog (bool, optional): 是否对 Y 轴使用对数坐标。
        sci (Tuple[float, float], optional): 用于设置x轴或y轴的科学计数法格式，格式为 (x轴缩放因子，y轴缩放因子)。
        color_group (List[int], optional): 每条曲线的颜色索引（用于自定义颜色组）。
        show (bool, optional): 是否在绘制完成后显示图像，默认不显示。
        save (bool, optional): 是否保存图像，默认不保存。
        font_style (str, optional): 字体样式，默认为 Times。可选值为 'sans' 或 None。
        font_weight (str, optional): 字体粗细，默认为 "bold"。可选值为 'bold' 或 'normal'。

    Returns:
        None. 图像将自动保存至默认目录，并在终端打印保存路径。
    """
    
    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    if show_residual:
        dpi, figuresize, savedir = set_residual_dpi_figsize_savedir()
        fig, (ax_main, ax_res) = plt.subplots(2, 1, figsize=figuresize, dpi=dpi,
                                            gridspec_kw={'height_ratios': [3, 1]}, sharex=True)
    else:
        dpi, figuresize, savedir = set_default_dpi_figsize_savedir()
        fig, ax_main = plt.subplots(figsize=figuresize, dpi=dpi)
        ax_res = None
    save_dir = os.path.join(savedir, "Curves")

    # Set the main axis style
    for ax in [ax_main] + ([ax_res] if ax_res is not None else []):
        set_ax_style(ax)
        ax.tick_params(axis='both', direction='in', width=3, which='both', pad=10)

    time_step = time_step or [0] * len(path)
    use_marker = use_marker or [False] * len(path)
    use_scatter = use_scatter or [False] * len(path)

    curves = []
    if len(path) > 1:
        if not xy_label or not xy_label[0] or not xy_label[1]:
            raise ValueError("When plotting multiple curves, please specify 'xy_label' explicitly.")
        ax_main.set_title(title_figure or f'Comparison of {xy_label[1]}', pad=20, fontweight=font_weight)
        
        for i in range(len(path)):
            x_data, y_data, x_colname, y_colname = load_data_csv(
                source=path[i],
                x_index=x[i] if isinstance(path[i], str) else None,
                y_index=y[i] if isinstance(path[i], str) else None,
                factor=factor[i] if factor else [(1.0, 0.0), (1.0, 0.0)],
                time_step=time_step[i]
            )
            curves.append((x_data, y_data))

            color_index = color_group[i] if color_group else i

            if use_scatter[i]:
                plot_scatter_style(ax_main, x_data, y_data, label[i], color_index)
            elif use_marker[i]:
                update_curve_plotting_with_styles(ax_main, x_data, y_data, label[i], color_index)
            else:
                ax_main.plot(x_data, y_data, label=label[i], linewidth=3,
                            color=GLOBAL_COLORS[color_index % len(GLOBAL_COLORS)])
    else:
        x_data, y_data, x_colname, y_colname = load_data_csv(
            source=path[0],
            x_index=x[0] if isinstance(path[0], str) else None,
            y_index=y[0] if isinstance(path[0], str) else None,
            factor=factor[i] if factor else [(1.0, 0.0), (1.0, 0.0)],
            time_step=time_step[0]
        )
        curves.append((x_data, y_data))

        color_index = color_group[0] if color_group else 10

        if use_scatter[0]:
            plot_scatter_style(ax_main, x_data, y_data, label[i], color_index)
        elif use_marker[0]:
            update_curve_plotting_with_styles(ax_main, x_data, y_data, label[i], color_index)
        else:
            ax_main.plot(x_data, y_data, label=label[0], linewidth=3, 
                         color=GLOBAL_COLORS[color_index % len(GLOBAL_COLORS)])
        
        if not x_colname or not y_colname:
            raise ValueError("The current CSV data does not have column names. Please specify the 'xy_label' parameter.")
        if not xy_label:
            xy_label = [x_colname, y_colname]
        ax_main.set_title(title_figure or f'Curve of {xy_label[1]}', pad=20, fontweight=font_weight)

    ax_main.set_xlabel(xy_label[0], fontweight=font_weight)
    ax_main.set_ylabel(xy_label[1], fontweight=font_weight)

    label_suffix = f"({information})" if information else None

    if sci[0]:
        apply_axis_scientific_format(ax_main, 'x', sci[0])
    if sci[1]:
        apply_axis_scientific_format(ax_main, 'y', sci[1])

    apply_axis_limits_and_ticks(
        ax=ax_main,
        curves=curves,
        xlim=xlim,
        ylim=ylim,
        tick_interval_x=tick_interval_x,
        tick_interval_y=tick_interval_y
    )

    if ylog:
        ax_main.set_yscale('log')
        ax_main.set_ylim(1e-3, 0.6)
        ax_main.yaxis.set_major_locator(FixedLocator([0.002, 0.01, 0.1, 0.6]))
        ax_main.yaxis.set_major_formatter(FuncFormatter(lambda y, _: f'{y:g}'))
        ax_main.yaxis.set_minor_locator(NullLocator())
        ax_main.yaxis.set_minor_formatter(NullFormatter())

    save_or_display_legend(
        ax=ax_main,
        save_dir=save_dir,
        figure_name_suffix=label_suffix,
        split_legend=split_legend,
        legend_location=legend_location,
        legend_ncol=legend_ncol or 1,
        dpi=dpi,
        xy_label=xy_label,
        show_legend=show_legend
    )

    if show_residual and len(curves) >= 2:
        plot_residual_curves(
            ax_res=ax_res,
            curves=curves,
            label=label,
            xy_label=xy_label,
            x_title_fallback=xy_label[0]
        )

    plt.tight_layout() 
    filename = generate_plot_filename(title=title_figure, suffix=label_suffix)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close()
    return save_path


def plot_curves(
    data: list[tuple[np.ndarray, np.ndarray]],
    label: list[str],
    information: str = None,
    factor: list[tuple[tuple, tuple]] = [None],
    time_step: list[int] = None,
    xy_label: tuple[str, str] = None,
    use_marker: list[bool] = None,
    use_scatter: list[bool] = None,
    tick_interval_x: float = None,
    tick_interval_y: float = None,
    legend_location: str = None,
    xlim: tuple[float, float] = None,
    ylim: tuple[float, float] = None,
    highlight_x: float = None,
    split_legend: bool = False,
    show_legend: bool = True,
    show_residual: bool = False,
    title_figure: str = None,
    legend_ncol: int = None,
    ylog: bool = False,
    sci: tuple[float, float] = (None, None),
    color_group: list[int] = None,
    show: bool = False,
    save: bool = False,
    font_style: str = None,
    font_weight: str = "bold",
) -> str:
    """
    绘制一条或多条曲线，支持多种样式控制与残差分析的通用函数。

    Args:
        data (List[Tuple[np.ndarray, np.ndarray]]): 每条曲线的数据，格式为 [(x_data, y_data), ...]。
        label (List[str]): 每条曲线的图例标签。
        information (str, optional): 文件名后缀信息，用于区分保存图像。
        factor (List[tuple[tuple, tuple]]], optional): x轴和y轴的缩放因子和偏移量，例如用于单位换算。格式为 [((x_scale, x_offset), (y_scale, y_offset)), ...]。默认均为 1 和 0。
        time_step (List[int], optional): 用于截断数据的步数，若为 0 表示不截断。
        xy_label (Tuple[str, str], optional): X 和 Y 轴标签，例如 ("Time (s)", "Stress (MPa)")。
        use_marker (List[bool], optional): 是否为每条曲线使用线+标记风格。
        use_scatter (List[bool], optional): 是否将每条曲线绘制为散点图。
        tick_interval_x (float, optional): X 轴主刻度间隔。
        tick_interval_y (float, optional): Y 轴主刻度间隔。
        legend_location (str, optional): 图例位置（如 'best'、'upper right' 等）。
        xlim (Tuple[float, float], optional): X 轴显示范围。
        ylim (Tuple[float, float], optional): Y 轴显示范围。
        highlight_x (float, optional): 保留接口，尚未启用，用于高亮指定 X 点。
        split_legend (bool, optional): 是否将图例单独绘制成图像保存。
        show_legend (bool, optional): 是否显示图例，默认显示。
        show_residual (bool, optional): 是否绘制与参考曲线（第 1 条）相比的残差图。
        title_figure (str, optional): 图像标题，也作为保存文件名的前缀。
        legend_ncol (int, optional): 图例列数，默认自动调整。
        ylog (bool, optional): 是否对 Y 轴使用对数坐标。
        sci (Tuple[float, float], optional): 用于设置x轴或y轴的科学计数法格式，格式为 (x轴缩放因子，y轴缩放因子)。
        color_group (List[int], optional): 每条曲线的颜色索引（用于自定义颜色组）。
        show (bool, optional): 是否在绘制完成后显示图像，默认不显示。
        save (bool, optional): 是否保存图像，默认不保存。
        font_style (str, optional): 字体样式，默认为 Times。可选值为 'sans' 或 None。
        font_weight (str, optional): 字体粗细，默认为 "bold"。可选值为 'bold' 或 'normal'.

    Returns:
        str: 保存的图像路径。

    Raises:
        ValueError: 如果数据格式不正确或缺少必要参数。
        ValueError: 如果 xy_label 在多条曲线绘制时未指定。
        ValueError: 如果 xy_label 在单条曲线绘制时未指定且数据没有列名。
    """
    
    if not font_style:
        if font_weight == 'bold':
            set_default_style(bold=True)
        elif font_weight == 'normal':
            set_default_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    elif font_style == 'sans':
        if font_weight == 'bold':
            set_sans_style(bold=True)
        elif font_weight == 'normal':
            set_sans_style(bold=False)
        else:
            raise ValueError("Invalid font_weight. Choose 'bold' or 'normal'.")
    else:
        raise ValueError("Invalid font_style. Choose 'sans' or None.")

    if show_residual:
        dpi, figuresize, savedir = set_residual_dpi_figsize_savedir()
        fig, (ax_main, ax_res) = plt.subplots(2, 1, figsize=figuresize, dpi=dpi,
                                              gridspec_kw={'height_ratios': [3, 1]}, sharex=True)
    else:
        dpi, figuresize, savedir = set_default_dpi_figsize_savedir()
        fig, ax_main = plt.subplots(figsize=figuresize, dpi=dpi)
        ax_res = None
    save_dir = os.path.join(savedir, "Curves")

    for ax in [ax_main] + ([ax_res] if ax_res is not None else []):
        set_ax_style(ax)
        ax.tick_params(axis='both', direction='in', width=3, which='both', pad=10)

    time_step = time_step or [0] * len(data)
    use_marker = use_marker or [False] * len(data)
    use_scatter = use_scatter or [False] * len(data)

    label_suffix = f"({information})" if information else None
        
    curves = []
    if len(data) > 1:
        if not xy_label or not xy_label[0] or not xy_label[1]:
            raise ValueError("When plotting multiple curves, please specify 'xy_label' explicitly.")

        for i in range(len(data)):
            if factor[i] is None:
                factor[i] = [(1.0, 0.0), (1.0, 0.0)]
            x_data, y_data = data[i]
            x_scale, x_offset = factor[i][0]
            y_scale, y_offset = factor[i][1]
            x_data = x_data * x_scale + x_offset
            y_data = y_data * y_scale + y_offset
            if time_step[i]:
                x_data = x_data[:time_step[i]]
                y_data = y_data[:time_step[i]]
            curves.append((x_data, y_data))

            color_index = color_group[i] if color_group else i

            if use_scatter[i]:
                plot_scatter_style(ax_main, x_data, y_data, label[i], color_index)
            elif use_marker[i]:
                update_curve_plotting_with_styles(ax_main, x_data, y_data, label[i], color_index)
            else:
                ax_main.plot(x_data, y_data, label=label[i], linewidth=3,
                            color=GLOBAL_COLORS[color_index % len(GLOBAL_COLORS)])
        
        ax_main.set_title(title_figure or f'Comparison of {xy_label[1]}', pad=20, fontweight=font_weight)
    
    else:
        i = 0
        x_data, y_data = data[i]
        if factor[i] is None:
            factor[i] = [(1.0, 0.0), (1.0, 0.0)]
        x_scale, x_offset = factor[i][0]
        y_scale, y_offset = factor[i][1]
        x_data = x_data * x_scale + x_offset
        y_data = y_data * y_scale + y_offset
        if time_step[i]:
            x_data = x_data[:time_step[i]]
            y_data = y_data[:time_step[i]]
        curves.append((x_data, y_data))     
        color_index = color_group[i] if color_group else 10

        if use_scatter[i]:
            plot_scatter_style(ax_main, x_data, y_data, label[i], color_index)
        elif use_marker[i]:
            update_curve_plotting_with_styles(ax_main, x_data, y_data, label[i], color_index)
        else:
            ax_main.plot(x_data, y_data, label=label[i], linewidth=3,
                        color=GLOBAL_COLORS[color_index % len(GLOBAL_COLORS)])
        
        if not xy_label:
            raise ValueError("For single curve, 'xy_label' is required when no column name is provided.")   
        ax_main.set_title(title_figure or f'Curve of {xy_label[1]}', pad=20, fontweight=font_weight)

    ax_main.set_xlabel(xy_label[0], fontweight=font_weight)
    ax_main.set_ylabel(xy_label[1], fontweight=font_weight)

    if sci[0]:
        apply_axis_scientific_format(ax_main, 'x', sci[0])
    if sci[1]:
        apply_axis_scientific_format(ax_main, 'y', sci[1])

    apply_axis_limits_and_ticks(
        ax=ax_main,
        curves=curves,
        xlim=xlim,
        ylim=ylim,
        tick_interval_x=tick_interval_x,
        tick_interval_y=tick_interval_y
    )

    if ylog:
        ax_main.set_yscale('log')
        ax_main.set_ylim(0.0, 0.6)
        ax_main.yaxis.set_major_locator(FixedLocator([0.002, 0.01, 0.1, 0.6]))
        ax_main.yaxis.set_major_formatter(FuncFormatter(lambda y, _: f'{y:g}'))
        ax_main.yaxis.set_minor_locator(NullLocator())
        ax_main.yaxis.set_minor_formatter(NullFormatter())

    save_or_display_legend(
        ax=ax_main,
        save_dir=save_dir,
        figure_name_suffix=label_suffix,
        split_legend=split_legend,
        legend_location=legend_location,
        legend_ncol=legend_ncol or 1,
        dpi=dpi,
        xy_label=xy_label,
        show_legend=show_legend
    )

    if show_residual and len(curves) >= 2:
        plot_residual_curves(
            ax_res=ax_res,
            curves=curves,
            label=label,
            xy_label=xy_label,
            x_title_fallback=xy_label[0]
        )

    plt.tight_layout()
    filename = generate_plot_filename(title=title_figure, suffix=label_suffix)
    save_path = os.path.join(save_dir, filename)
    if save:
        save_figure(save_path, dpi=dpi)
    if show:
        plt.show()
    plt.close()
    return save_path



'''
Author: bo-qian bqian@shu.edu.cn
Date: 2025-06-25 15:29:33
LastEditors: bo-qian bqian@shu.edu.cn
LastEditTime: 2025-08-28 20:22:24
FilePath: /boviz/src/boviz/style.py
Description: This module provides functions to set default styles for plots in boviz.
Copyright (c) 2025 by Bo Qian, All Rights Reserved. 
'''

import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from matplotlib import ticker

def set_default_style(bold: bool = True):
    plt.rcParams.update({
        'font.family': 'serif',
        'font.serif': ['Times New Roman'],
        'font.weight': 'bold' if bold else 'normal',
        'font.size': 32,
        'mathtext.fontset': 'custom',
        'mathtext.rm': 'Times New Roman',
        'mathtext.it': 'Times New Roman:italic',
        'mathtext.bf': 'Times New Roman:bold' if bold else 'Times New Roman',
        'axes.unicode_minus': False,
        'legend.fontsize': 28
    })


def set_sans_style(bold: bool = True):
    """
    Set the default style for plots using sans-serif fonts.
    """
    plt.rcParams.update({
        'font.family': 'sans-serif',
        'font.sans-serif': ['Arial'],
        'font.weight': 'bold' if bold else 'normal',
        'font.size': 32,
        'mathtext.fontset': 'custom',
        'mathtext.rm': 'Arial',
        'mathtext.it': 'Arial:italic',
        'mathtext.bf': 'Arial:bold' if bold else 'Arial',
        'axes.unicode_minus': False,
        'legend.fontsize': 28
    })


def set_ax_style(ax, linewidth=3):
    """
    Set the style for a given Axes object.

    Args:
        ax (matplotlib.axes.Axes): The Axes object to style.
    Returns:
        None
    """
    ax.spines['top'].set_linewidth(linewidth)
    ax.spines['bottom'].set_linewidth(linewidth)
    ax.spines['left'].set_linewidth(linewidth)
    ax.spines['right'].set_linewidth(linewidth)


def apply_axis_scientific_format(ax, axis: str, scale: float, fontsize: int = 22):
    """
    将某个轴设置为科学计数格式。

    Args:
        ax: matplotlib 坐标轴对象。
        axis (str): 'x' 或 'y'。
        scale (float): 缩放因子。
        fontsize (int): 标注字号。
    Returns:
        None
    """
    if not scale or scale == 1.0:
        return
    sci_power = int(np.log10(scale))
    formatter = FuncFormatter(lambda val, _: f"{val / scale:.1f}")

    if axis == 'y':
        ax.yaxis.set_major_formatter(formatter)
        ax.text(0.01, 1.01, rf'$\times10^{{{sci_power}}}$',
                transform=ax.transAxes,
                fontsize=fontsize,
                fontweight='bold',
                verticalalignment='bottom',
                horizontalalignment='left')
    elif axis == 'x':
        ax.xaxis.set_major_formatter(formatter)
        ax.text(1.01, -0.13, rf'$\times10^{{{sci_power}}}$',
                transform=ax.transAxes,
                fontsize=fontsize,
                fontweight='bold',
                verticalalignment='top',
                horizontalalignment='right')
        

def apply_axis_limits_and_ticks(
    ax: plt.Axes,
    curves: list[tuple[np.ndarray, np.ndarray]],
    xlim: tuple[float, float] = None,
    ylim: tuple[float, float] = None,
    tick_interval_x: float = None,
    tick_interval_y: float = None
):
    """
    设置 x/y 轴的显示范围与主刻度间隔。

    Args:
        ax: matplotlib Axes 对象。
        curves: 曲线数据列表，每项为 (x_data, y_data)。
        xlim: x 轴范围，如 (0, 10)。若为 None，则自动判断。
        ylim: y 轴范围，如 (-1, 1)。若为 None，则自动判断。
        tick_interval_x: x 轴主刻度间隔。
        tick_interval_y: y 轴主刻度间隔。
    Returns:
        None
    """
    # ----- X 轴主刻度 -----
    if tick_interval_x:
        x_min = min(min(c[0]) for c in curves)
        x_max = max(max(c[0]) for c in curves)
        ax.set_xticks(np.arange(x_min, x_max + tick_interval_x, tick_interval_x))

    # ----- Y 轴主刻度（先粗设）-----
    if tick_interval_y:
        y_min = min(min(c[1]) for c in curves)
        y_max = max(max(c[1]) for c in curves)
        ax.set_yticks(np.arange(y_min, y_max + tick_interval_y, tick_interval_y))

    # ----- X/Y 轴显示范围 -----
    if xlim:
        ax.set_xlim(*xlim)
    if ylim:
        ax.set_ylim(*ylim)
        # 若设定了 ylim 且 tick_interval_y 也存在，则精确地重新设定 y ticks
        if tick_interval_y:
            ax.set_yticks(np.arange(ylim[0], ylim[1] + 0.1 * tick_interval_y, tick_interval_y))


def save_or_display_legend(
    ax,
    save_dir: str,
    figure_name_suffix: str,
    split_legend: bool = False,
    legend_location: str = 'best',
    legend_ncol: int = 1,
    dpi: int = 300,
    xy_label: list[str, str] = None,
    show_legend: bool = True
):
    """
    处理图例显示或保存为单独图像的逻辑。

    Args:
        ax: 主图坐标轴。
        save_dir: 图像保存目录。
        figure_name_suffix: 用于生成图例文件名的后缀。
        split_legend: 是否将图例单独绘制保存。
        legend_location: 图例位置（默认 'best'）。
        legend_ncol: 图例列数。
        dpi: 图像保存分辨率。
        y_label: 用于命名的 y 轴标签。
    """
    if split_legend:
        legend = ax.legend(fontsize='small', loc=legend_location or 'best')
        fig_leg = plt.figure(figsize=(4, 2))
        fig_leg.legend(*ax.get_legend_handles_labels(), loc='center')
        fig_leg.tight_layout()

        if xy_label:
            y_label = xy_label[1]
        else:
            ValueError("xy_label must be provided.")
        legend_path = os.path.join(save_dir, f'Legend of {y_label}{figure_name_suffix}.png')
        fig_leg.savefig(legend_path, dpi=dpi, bbox_inches='tight')
        plt.close(fig_leg)
        legend.remove()
    elif show_legend:
        ax.legend(fontsize='small', loc=legend_location or 'best', ncol=legend_ncol or 1)


def plot_residual_curves(
    ax_res,
    curves: list[tuple[np.ndarray, np.ndarray]],
    label: list[str],
    xy_label: tuple[str, str] = None,
    x_title_fallback: str = None
):
    """
    绘制残差图：将第 1 条曲线作为参考，绘制其与其他曲线之间的差值。

    Args:
        ax_res: matplotlib 的残差子图坐标轴。
        curves: 所有曲线的 (x, y) 数据对。
        label: 每条曲线的标签。
        xy_label: 坐标轴标签元组，用于设置 x 轴名。
        x_title_fallback: 如果 xy_label 不存在时使用的 x 轴名。
    """
    if len(curves) < 2:
        return

    x_ref, y_ref = curves[0]
    for i in range(1, len(curves)):
        x_i, y_i = curves[i]
        y_interp = np.interp(x_ref, x_i, y_i)
        residual = y_interp - y_ref
        ax_res.plot(x_ref, residual,
                    label=f'{label[i]} - {label[0]}', linewidth=2)

    # 设置 y 轴以 ×10⁻² 显示
    ax_res.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x, _: f'{x * 1e2:.1f}'))
    ax_res.text(0.01, 1.02, r'$\times10^{-2}$', transform=ax_res.transAxes,
                fontsize=20, verticalalignment='bottom', horizontalalignment='left')

    ax_res.axhline(0, color='gray', linestyle='--', linewidth=1)
    ax_res.set_ylabel("Residual", fontweight='bold')
    ax_res.set_xlabel(xy_label[0] if xy_label and xy_label[0] else x_title_fallback, fontweight='bold')
    ax_res.legend(fontsize=20)


def set_smart_xy_ticks(ax, extent=None):
    """
    为 ax 自动设置 x 和 y 轴的主刻度，使其间隔为 (5,10,20,25,30,50,100) 中的一个，
    且总刻度数在 5 到 9 之间。
    
    参数：
        ax : matplotlib.axes.Axes
            目标坐标轴对象
        extent : tuple or None
            可选的 (x0, x1, y0, y1)，若为 None 则自动从 ax 获取当前坐标范围
    """
    step_candidates = (5, 10, 20, 25, 30, 40, 50, 100)

    def compute_ticks(start, end):
        range_ = end - start
        for step in step_candidates:
            ticks_count = range_ / step
            if ticks_count.is_integer() and 4 <= ticks_count <= 7:
                break
        else:
            # fallback：选一个最接近7个刻度的
            step = min(step_candidates, key=lambda s: abs((range_ / s) - 6))

        ticks = np.arange(start, end + 0.5 * step, step)
        return np.round(ticks, 8)

    if extent is not None:
        x0, x1, y0, y1 = extent
    else:
        x0, x1 = ax.get_xlim()
        y0, y1 = ax.get_ylim()

    ax.set_xticks(compute_ticks(x0, x1))
    ax.set_yticks(compute_ticks(y0, y1))
    ax.set_xlim(x0, x1)
    ax.set_ylim(y0, y1)



"""
boviz: A Python plotting toolkit for scientific visualization.

This package exports:
    - plot_curves: Plotting of scientific curves with customizable styles.
    - plot_initial_particle_schematic: Visualization of initial particle schematics.
    - set_default_style: Set global matplotlib style for consistent appearance.
    - generate_plot_filename: Utility for generating plot filenames.
    - GLOBAL_COLORS, DEFAULT_SAVE_DIR, DEFAULT_DPI, DEFAULT_FIGSIZE: Global configuration constants.

Programs that import and use 'boviz' can easily create publication-quality plots
with consistent styles and convenient utilities.

Author: Bo Qian
Email: bqian@shu.edu.cn
"""

from .config import GLOBAL_COLORS, set_default_dpi_figsize_savedir, set_residual_dpi_figsize_savedir
from .style import set_default_style, set_ax_style, apply_axis_scientific_format, apply_axis_limits_and_ticks, save_or_display_legend, plot_residual_curves, set_sans_style, set_smart_xy_ticks
from .utils import generate_plot_filename, load_data_csv, save_figure, generate_particle_layout, build_tanh_phase_field, load_exodus_data, get_math_label, _broadcast, _bbox_cols_from_gridspec, _bbox_rows_from_axes
from .curves import plot_curves_csv, plot_curves
from .histogram import plot_histogram
from .heatmap import plot_heatmap_particle, plot_heatmap_exodus2d, plot_heatmap_exodus2d_grid
from .schematic import plot_initial_particle_schematic